/*
   Copyright 2011 Evan Williams

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 */

package uk.ac.reading.evanWilliams.javaResources.dataStructures;

import java.util.ArrayList;
import java.util.Arrays;

/**
 * Heap structure for generic data items.
 * 
 * @author Evan Williams<br />
 *         evan.williams@bcs.org<br />
 *         {@linkplain http://my-technicolour-dreams.com}
 * 
 * @param <T>
 *            The data type that will be stored in the heap. Must implement the
 *            comparable interface with itself
 */
public class Heap<T extends Comparable<? super T>> {
	public static void main(String[] args) {
		// Integer[] l = { 41, 24, 31, 66, 4, 65 };
		// Heap<Integer> a = new Heap<Integer>(false, l);
		// Heap<Integer> b = new Heap<Integer>(false, l);
		// Heap<Integer> c = new Heap<Integer>(false, l);
		//
		// System.out.println("Original list:   " + a.toString());
		//
		// a.insertRemake(17);
		// b.insertTopDown(17);
		// c.insertBottomUp(17);
		//
		// System.out.println("Expected result: " + a.toString());
		// System.out.println("TD Result:       " + b.toString());
		// System.out.println("BU Result:       " + c.toString());
		//

		// Create a list of vectors
		// ArrayList<Vector> vectorList = new ArrayList<Vector>();
		// vectorList.add(new Vector(1, 0, 1));
		// vectorList.add(new Vector(2, 0, 3));
		// vectorList.add(new Vector(4, 1, 2));
		// vectorList.add(new Vector(3, 1, 4));
		// vectorList.add(new Vector(7, 2, 5));
		// vectorList.add(new Vector(5, 3, 4));
		// vectorList.add(new Vector(2, 4, 6));
		// vectorList.add(new Vector(9, 4, 8));
		// vectorList.add(new Vector(2, 5, 4));
		// vectorList.add(new Vector(1, 6, 7));
		// vectorList.add(new Vector(4, 8, 7));
		//
		// //Build a heap with this list in it initially
		// Heap<Vector> vHeap = new Heap<Vector>(false,
		// vectorList.toArray(new Vector[vectorList.size()]));
		// //Output the current state of the heap
		// System.out.println(vHeap.toString());
		//
		// //Add another vector to the heap
		// vHeap.insertBottomUp(new Vector(3,0,4));
		// //Output the current state of the heap
		// System.out.println(vHeap.toString());

		// A list of 50 randomly generated integers to use as input.
		Integer[] l = { 22, 63, 5, 53, 38, 71, 93, 46, 43, 21, 60, 92, 49, 98,
				34, 58, 100, 70, 35, 89, 11, 10, 18, 33, 87, 51, 27, 83, 32,
				24, 15, 19, 65, 37, 90, 55, 66, 79, 88, 8, 28, 85, 2, 31, 4,
				97, 72, 7, 67, 75, 47 };
		// Integer[] l = { 73, 99, 77, 67, 86, 65, 98, 33, 80, 44, 43, 20, 37,
		// 7,
		// 30, 45, 59, 45, 45, 36, 1, 83, 41, 13, 89, 70, 10, 20, 73, 7,
		// 99, 59, 6, 17, 54, 95, 59, 11, 62, 56, 73, 91, 55, 74, 95, 38,
		// 46, 93, 40, 94 };
		// Integer[] l = { 97, 87, 69, 51, 64, 71, 2, 69, 34, 22, 66, 96, 41,
		// 95,
		// 83, 29, 57, 5, 67, 58, 77, 23, 49, 28, 42, 74, 78, 100, 68, 29,
		// 82, 16, 49, 50, 6, 37, 46, 46, 40, 60, 34, 13, 81, 78, 11, 90,
		// 92, 91, 93, 87 };
		// Integer[] l = { 47, 87, 2, 55, 43, 100, 18, 89, 7, 34, 88, 19, 28,
		// 30,
		// 77, 66, 73, 63, 5, 93, 62, 17, 74, 55, 5, 47, 71, 73, 64, 10,
		// 23, 87, 70, 85, 32, 18, 21, 52, 78, 75, 76, 46, 53, 96, 94, 3,
		// 50, 88, 89, 42 };
		// Integer[] l = { 93, 38, 80, 41, 57, 66, 11, 31, 92, 30, 55, 30, 62,
		// 65,
		// 23, 23, 72, 53, 33, 82, 69, 26, 84, 37, 81, 86, 14, 35, 29, 1,
		// 42, 78, 91, 35, 100, 42, 51, 24, 48, 12, 62, 79, 26, 39, 95,
		// 56, 19, 35, 63, 27 };
		// Integer[] l = { 56, 100, 45, 33, 23, 31, 15, 92, 8, 75, 58, 53, 34,
		// 56,
		// 66, 65, 88, 80, 12, 26, 96, 46, 54, 48, 1, 70, 43, 18, 81, 56,
		// 96, 1, 49, 31, 50, 90, 96, 65, 39, 60, 27, 1, 1, 70, 90, 42, 9,
		// 85, 95, 59 };
		// Integer[] l = { 34, 60, 93, 57, 40, 1, 91, 9, 28, 99, 70, 93, 74, 99,
		// 43, 47, 55, 95, 23, 38, 95, 8, 16, 82, 17, 27, 94, 79, 64, 26,
		// 74, 64, 68, 47, 19, 7, 40, 38, 17, 100, 23, 41, 99, 76, 21, 72,
		// 100, 93, 73, 26 };
		// Integer[] l = { 46, 4, 24, 72, 93, 19, 89, 100, 88, 55, 8, 83, 6, 7,
		// 47, 79, 66, 64, 46, 60, 91, 12, 86, 58, 64, 65, 79, 40, 45, 99,
		// 3, 1, 87, 4, 7, 9, 29, 7, 11, 62, 96, 18, 93, 40, 66, 62, 28,
		// 37, 41, 43 };
		// Integer[] l = { 41, 40, 31, 50, 12, 14, 52, 7, 59, 14, 67, 70, 4, 5,
		// 67, 70, 30, 67, 35, 67, 42, 91, 16, 27, 85, 53, 63, 17, 57, 26,
		// 63, 95, 52, 1, 95, 80, 6, 58, 43, 72, 66, 43, 86, 13, 74, 73,
		// 76, 77, 34, 23 };
		// Integer[] l = { 52, 94, 35, 8, 96, 32, 84, 97, 22, 41, 15, 2, 53, 4,
		// 14, 80, 76, 50, 71, 89, 68, 10, 66, 69, 55, 35, 19, 60, 5, 36,
		// 82, 45, 17, 3, 83, 27, 11, 33, 19, 93, 57, 54, 57, 70, 33, 20,
		// 11, 88, 69, 71 };
		// A second list that will contain a selection from the main list
		Integer[] sub;

		Heap<Integer> insert, heapify;
		String out = "";

		// Create a title row for the output
		// System.out
		//
		// .println("Input Size, Number of loops for Insert, Number of loops for Heapify");
		//
		for (int i = 0; i < 50; i++) {
			// Get the first i items from the number list
			sub = Arrays.copyOfRange(l, 0, i);

			// Create a new heap, with the sublist inserted by the heapify
			// method
			heapify = new Heap<Integer>(false, sub);

			// Create a new heap, with the sublist inserted by the insert method
			insert = new Heap<Integer>(false);
			for (int j : sub) {
				insert.insertBottomUp(j);
			}

			out = i + "," + insert.loops + "," + heapify.loops;
			System.out.println(out);
		}
	}

	private ArrayList<T> heap; // The 1D array the data is stored in

	private int maximise; // Used to decide sorting order. Multiplied against

	// Comparisons, meaning a negative value will invert the
	// result of a comparison

	public int loops;

	/**
	 * Constructor for creating an empty Heap of generic objects.
	 * 
	 * @param m
	 *            Whether the heap should be sorted high-to-low, or low-to-high.
	 *            True for high-to-low
	 */
	public Heap(boolean m) {
		heap = new ArrayList<T>(); // Initialise the array list
		maximise = (m ? 1 : -1); // If the user specifies high-to low, provide
									// positive number, otherwise provide a
									// negative
		loops = 0;
	}

	/**
	 * Constructor for creating a Heap of generic objects, containing all the
	 * items in the array init
	 * 
	 * @param init
	 *            An array of generic objects to be placed in to the Heap.
	 * @param m
	 *            Whether the heap should be sorted high-to-low, or low-to-high
	 */
	public Heap(boolean m, T[] init) {
		heap = new ArrayList<T>(); // Initialise the array list
		maximise = (m ? 1 : -1); // If the user specifies high-to low, provide
									// positive number, otherwise provide a
									// negative
		loops = 0;

		for (T i : init)
			// Loop through the provided array
			heap.add(i); // Add each item to the heap
		makeHeap(); // Fix the heap
	}

	/**
	 * Compares a root node to its child.
	 * 
	 * @param root
	 *            The index of the root node
	 * @param rchild
	 *            Whether it was the right-hand child. False for left child,
	 *            true for right child
	 * @return True if the items are in order
	 */
	private boolean comp(int root, boolean rchild) {
		// Form the index of the selected child
		int child = getChild(root, rchild);
		// If the chosen child does not exist (as would be the case for the
		// right child of the final parent if the total number of nodes is even)
		// return that they are in order
		if (child >= heap.size())
			return true;

		// Compare the item to the specified child, then multiply by the
		// maximise variable to decide whether they're in order. An overall
		// negative result indicates that they are.
		int comp = heap.get(root).compareTo(heap.get(child)) * maximise;
		if (comp < 0) // If the items are out of order
			return false; // Tell the calling method that they are
		else
			// Otherwise
			return true; // Tell the calling method they are in order
	}

	/**
	 * Fetches the index of a child of the specified node. Does not check
	 * whether the child actually exists.
	 * 
	 * @param parent
	 *            The node to find the children of
	 * @param isRightChild
	 *            Whether you are fetching the right hand child
	 * @return The location of the child.
	 */
	private int getChild(int parent, boolean isRightChild) {
		return 2 * parent + (isRightChild ? 2 : 1);
	}

	/**
	 * @return The top item of the Heap
	 */
	public T getTop() {
		return heap.get(0);
	}

	/**
	 * Finds how many children a given node has
	 * 
	 * @param parent
	 *            The node to examine
	 * @return The number of children owned by that node
	 */
	private int hasChildren(int parent) {
		// Find the relative position of where the right child would be compared
		// to the end of the heap
		int comp = getChild(parent, true) - heap.size();
		if (comp < 0) // If both children exist
			return 2;
		if (comp == 0) // If only the left child exists
			return 1;
		// If neither child exists
		return 0;
	}

	/**
	 * Applies the Heap property to the sub-tree beginning with a specified node
	 * 
	 * @param parent
	 *            The top of the tree to work with
	 */
	private void heapify(int parent) {
		boolean largest = rightHigher(parent);
		boolean swap = false;
		final boolean[] bvals = { largest, !largest };
		int numChildren = hasChildren(parent);
		loops++;

		// Check each child, assuming a swap has not been declared
		for (int i = 0; i < 2 && !swap; i++) {
			if (i < numChildren) { // If the child exists
				if (!comp(parent, bvals[i])) // If it is out of order
					swap = true; // Mark that a swap needs to occur
			}
		}

		if (swap) { // If a swap has been declared
			swap(parent, largest); // Swap with the higher priority child
			heapify(getChild(parent, largest)); // Call again for the sublist
		}
	}

	/**
	 * Adds a new item to the heap, ensuring they stay sorted by adding it to
	 * the bottom and letting it "float" up
	 * 
	 * @param node
	 *            The data to be added
	 */
	public void insertBottomUp(T node) {
		heap.add(node); // Add node to end of heap
		int i = heap.size() / 2 - 1; // Keep track of node's parent
		int j = heap.size() - 1; // Keep track of node

		boolean isRightChild;

		while (i >= 0) { // While i is within the bounds of the array
			loops++;
			isRightChild = (j % 2 == 0); // Check which child it is dealing with
			if (!comp(i, isRightChild)) { // if the node and child are out of
											// order
				swap(i, isRightChild); // swap them
				// move the markers for the node and parent up one level
				j = i;
				i = (i + 1) / 2 - 1;
			} else
				// Otherwise
				i = -1; // Break the loop
		}
	}

	/**
	 * Adds a new item to the heap, ensuring they stay in the correct order by
	 * rebuilding the heap every time
	 * 
	 * @param data
	 *            The item to be added to the Heap
	 * @return Whether it worked
	 */
	public boolean insertRemake(T data) {
		boolean worked = heap.add(data);
		if (worked) // If adding the data item to the array works
			makeHeap(); // Fix the heap

		return worked; // Return whether it worked
	}

	/**
	 * Adds a new item to the heap, ensuring they stay in the correct order by
	 * placing at the start of the array list, and letting it "sink" to its
	 * correct position
	 * 
	 * @param data
	 *            The item to be added to the Heap
	 */
	public void insertTopDown(T data) {
		heap.add(0, data); // Add the item to the top of the heap

		int size = heap.size();
		int i = 0;
		int j = 1;
		boolean rightFirst;

		while (j < size) { // While the inserted node has children
			rightFirst = rightHigher(i);
			// If it is of higher priority than the highest priority child
			if (!comp(i, rightFirst)) {
				swap(i, rightFirst); // Swap with that child

				// Mark which child was swapped with
				if (rightFirst)
					j++;
			}

			// Otherwise, if it is of higher priority than the other child
			else if (!comp(i, !rightFirst)) {
				swap(i, rightFirst); // Swap with the higher priority child

				// Mark which child was swapped with
				if (rightFirst)
					j++;
			} else
				j = size; // Escape the loop
			i = j; // Keep track of where the inserted node is currently
			j = 2 * i + 1; // Keep track of where the inserted node's children
							// are
		}
	}

	/**
	 * @return Whether the Heap has any components
	 */
	public boolean isEmpty() {
		return heap.isEmpty();
	}

	/**
	 * Rectifies the Heap, putting all of the items in the correct order
	 */
	private void makeHeap() {
		// Find the largest node with children. Rounding is toward 0 in integer
		// division, so it works for even and odd sized heaps
		int size = heap.size();
		int max = size / 2;

		// Work from the bottom up, applying the heap property to increasingly
		// larger, but already heaped trees
		for (int i = max - 1; i >= 0; i--)
			heapify(i);
	}

	/**
	 * Removes an object at a specific index, then fixes the heap
	 * 
	 * @param index
	 *            The location to remove from
	 * @return Whether the item was removed
	 */
	public boolean remove(int index) {
		if (heap.remove(index) != null) { // If removing the object works
			makeHeap(); // Fix the heap
			return true; // Return that it works
		} else
			// Otherwise
			return false; // return that it didn't work
	}

	/**
	 * Removes the first instance of the specific object from the heap, then
	 * fixes the heap
	 * 
	 * @param instance
	 *            The object to remove
	 * @return Whether the object was there, and therefore removed
	 */
	public boolean remove(T instance) {
		if (heap.remove(instance)) { // If removing the object worked
			makeHeap(); // Fix the heap
			return true; // Return that it worked
		} else
			// Otherwise
			return false; // Return that it didn't
	}

	/**
	 * Assesses which of the two children of a node is of higher priority. Will
	 * produce fallacious results if the right-hand node does not exist
	 * 
	 * @param root
	 *            The index of the node whose children are being checked
	 * @return <ul>
	 *         <li><strong>True</strong> if the right-hand node is of higher
	 *         priority.</li>
	 *         <li><strong>False</strong> if either the left-hand node is of
	 *         higher priority, or there are less than two children to the node</li>
	 *         </ul>
	 */
	private boolean rightHigher(int root) {
		// Calculate indices of child nodes
		int child1 = getChild(root, false);
		int child2 = getChild(root, true);

		if (child2 >= heap.size()) // If both children do no exist
			return false; // Report that the right child is not higher priority

		// Compare the two children
		int comp = heap.get(child1).compareTo(heap.get(child2)) * maximise;

		if (comp < 0) // If right is of higher priority
			return true; // Report this
		else
			// Otherwise
			return false; // Report that right child is not higher priority
	}

	/**
	 * Swaps the root node with with child indicated by the second parameter
	 * 
	 * @param root
	 *            The index of the root node
	 * @param rchild
	 * @return
	 */
	private void swap(int root, boolean rchild) {
		// Find the index of the specified child
		int child = getChild(root, rchild);
		T temp = heap.get(root); // Store the value of the main node
		heap.set(root, heap.get(child)); // Replace the main node with the child
											// node
		heap.set(child, temp); // Replace the child node with the parent node's
								// stored value
	}

	/**
	 * Outputs the content of the heap as a string. Uses the 1D array
	 * visualisation of a heap
	 */
	@Override
	public String toString() {
		String output = ""; // Create a string to store the output as the method
							// works
		for (T i : heap)
			// Loop through all the items in the heap
			output += i.toString() + ", "; // Add the string form of the current
											// object to the output string

		return output; // Return the output
	}
}
