/*
   Copyright 2011 Evan Williams

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 */

package uk.ac.reading.evanWilliams.javaResources.delta;

/**
 * <p>
 * An interface that should be implemented on any class that wishes to utilise
 * the Delta compression engine. In addition to the methods specified here, the
 * class should also have a series of inner classes - each of which extend
 * <code>Delta</code> - one for each type of change that can be applied to the
 * class. The Class should then create new objects of the class whenever a
 * change is to be applied to the object, and allow the <code>Delta</code> to
 * implement the change.
 * </p>
 * 
 * <p>
 * All changes should have a strict inverse operation, i.e. if one operation
 * moves the object from state a to state b, there should be another operation
 * to return the object to state a. The inverse operation should be described in
 * the inner <code>Delta</code>'s <code>revokeDelta()</code> method.
 * </p>
 * 
 * <p>
 * The exact implementation of this interface is a little odd, and probably
 * needs fixing, but when writing it I needed it to work quickly rather than
 * prettily. If anyone has a better way of writing this, please change it. As it
 * stands, you need to declare your class as
 * <code>ClassName implements DeltaTarget&lt;ClassName&gt;</code>
 * </p>
 * 
 * @author Evan Williams<br />
 *         evan.williams@bcs.org<br />
 *         {@linkplain http://my-technicolour-dreams.com}
 * 
 */
@SuppressWarnings("rawtypes")
public interface DeltaTarget<T extends DeltaTarget> {
    /**
     * Attaches all of the Deltas in the argument to the
     * 
     * @return
     */
    public DeltaQueue<T> attachAll(DeltaQueue<T> dl);

    /**
     * Takes a Delta attached to one object, and generates an identical Delta,
     * aside from it now being attached to this one.
     * 
     * @param d
     *            A Delta attached to another object
     * @return The new Delta, attached to this object
     */
    public Delta<T> attatchDelta(Delta<T> d);
}
