/*
   Copyright 2011 Evan Williams

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 */

package uk.ac.reading.evanWilliams.javaResources.delta;

import java.util.ArrayList;
import java.util.Date;

/**
 * An example class to demonstrate the functionality of the delta compression
 * engine
 * 
 * @author Evan Williams<br />
 *         evan.williams@bcs.org<br />
 *         {@linkplain http://my-technicolour-dreams.com}
 * 
 */
public class StringEdit implements DeltaTarget<StringEdit> {

    /**
     * A Delta that describes the action of injecting one string into another
     * 
     * @author Evan
     * 
     */
    public class StringInject extends Delta<StringEdit> {

	private final String midString;
	private final int startPoint;

	StringInject(final String m, final int s) {
	    super(StringEdit.this);
	    midString = m;
	    startPoint = s;
	}

	StringInject(final String m, final int s, final Date d) {
	    super(StringEdit.this, d);
	    midString = m;
	    startPoint = s;
	}

	@Override
	public Object[] getParams() {
	    final Object[] output = new Object[4];
	    output[0] = getTime();
	    output[1] = getTarget();
	    output[2] = startPoint;
	    output[3] = midString;
	    return output;
	}

	@Override
	public boolean invokeDelta() {
	    return stringInject(midString, startPoint);
	}

	@Override
	public boolean revokeDelta() {
	    return cutString(startPoint, startPoint + midString.length());
	}

    }

    public static void main(final String[] args) {
	// StringEdit testString = new StringEdit("abcdefghijklmnopqrstuvwxyz");
	// System.out.println(testString);
	// StringInject test = testString.new
	// StringInject("elpi'mstuckinanalphabetfactory", 8);
	// test.invokeDelta();
	// System.out.println(testString);
	// test.revokeDelta();
	// System.out.println(testString);

	// StringEdit testString = new StringEdit("abcdefghijklmnopqrstuvwxyz");
	// System.out.println(testString);
	//
	// ArrayList<StringInject> nums = new ArrayList<StringInject>();
	// for(int i = 0; i < 26; i++)
	// {
	// nums.add(testString.new StringInject(Integer.toString(i%10),2*i+1));
	// try {
	// Thread.sleep(10);
	// } catch (InterruptedException e) {
	// e.printStackTrace();
	// }
	// }

	// DeltaQueue dq = new DeltaQueue();
	// for(Delta d : nums){
	// dq.addDelta(d);
	// }
	//
	//
	// System.out.println(testString);
	// for(Delta d : nums){
	// dq.undo();
	// }
	// System.out.println(testString);
	//
	// int tmp;
	// while(!nums.isEmpty()){
	// while(true){
	// tmp = (int) (Math.random() * nums.size());
	// if(nums.get(tmp).startPoint <= testString.length())
	// break;
	// }
	// dq.addDelta(nums.get(tmp));
	// nums.remove(tmp);
	// }

	// DeltaQueue<StringEdit>[] dq = new DeltaQueue[4];
	//
	// dq[0] = new DeltaQueue<StringEdit>();
	// dq[1] = new DeltaQueue<StringEdit>();
	// dq[2] = new DeltaQueue<StringEdit>();
	// dq[3] = new DeltaQueue<StringEdit>();
	//
	// for(int i = 0; i < 13; i++)
	// {
	// dq[0].addDelta(nums.get(2*i));
	// dq[1].addDelta(nums.get(2*i+1));
	// dq[2].addDelta(nums.get(2*i));
	// dq[3].addDelta(nums.get(2*i+1));
	// }
	//
	// System.out.println(dq[0]);
	// System.out.println(dq[1]);
	//
	// dq[2].union(dq[1]);
	// dq[3].union(dq[0]);
	// dq[2].difference(dq[0]);
	// dq[3].difference(dq[1]);
	// System.out.println(dq[0]);
	// System.out.println(dq[1]);
	// System.out.println(dq[2]);
	// System.out.println(dq[3]);
	//
	// System.out.println(testString);
	//
	// System.out.println(nums.get(0).getClass());

	final StringEdit a = new StringEdit("abcdefghijklmnopqrstuvwxyz");
	final StringEdit b = new StringEdit("abcdefghijklmnopqrstuvwxyz");

	System.out.println(a + "\n" + b);

	a.deltas.addDelta(a.new StringInject("elpi'mstuck", 8));
	b.deltas.addDelta(b.new StringInject("inanalphebetfactory", 19));

	System.out.println(a + "\n" + b);

	a.deltas.addDelta(a.attatchDelta(b.deltas.undo()));

	System.out.println(a + "\n" + b);
	// }
	// String tmp = "elpi'mstuckinthealphabet";
	//
	// try {
	// for(int i = 0; i < tmp.length()/2 - 1;i++){
	// a.deltas.addDelta(a.new
	// StringInject(tmp.substring(2*i,2*i+1),8+2*i));
	// Thread.sleep(100);
	// b.deltas.addDelta(b.new
	// StringInject(tmp.substring(2*i+1,2*i+2),9+2*i));
	// Thread.sleep(100);
	// }
	// } catch (InterruptedException e) {
	//
	// e.printStackTrace();
	// }
	//
	// System.out.println(a + "\n" + b);
	//
	// a.deltas.union(a.attachAll(b.deltas));
	//
	// System.out.println(a + "\n" + b);
    }

    private String main;
    public DeltaQueue<StringEdit> deltas;

    public StringEdit() {
	main = "";
	deltas = new DeltaQueue<StringEdit>();
    }

    public StringEdit(final String s) {
	main = s;
	deltas = new DeltaQueue<StringEdit>();
    }

    @Override
    public DeltaQueue<StringEdit> attachAll(final DeltaQueue<StringEdit> dl) {

	final DeltaQueue<StringEdit> output = new DeltaQueue<StringEdit>();
	for (final Delta<StringEdit> d : dl)
	    output.addDelta(attatchDelta(d));

	return output;

    }

    @Override
    public Delta<StringEdit> attatchDelta(final Delta<StringEdit> d) {
	final ArrayList<String> innerClasses = new ArrayList<String>();
	innerClasses.add(this.new StringInject("a", 1).getClass().toString());

	switch (innerClasses.indexOf(d.getClass().toString())) {
	case 0:
	    return new StringInject((String) d.getParams()[3],
		    (Integer) d.getParams()[2], (Date) d.getParams()[0]);
	default:
	    return null;
	}

    }

    private boolean cutString(final int start, final int end) {
	main = main.substring(0, start) + main.substring(end);
	return true;
    }

    private boolean stringInject(final String m, final int s) {
	main = main.substring(0, s) + m + main.substring(s);
	return true;
    }

    @Override
    public String toString() {
	return main;
    }
}
